from marshmallow import Schema, post_load, pre_dump, validate
from marshmallow.fields import Field, String, Float, Boolean, Integer
from marshmallow.validate import OneOf, Range

BANDS = ["A", "B"]
BAND_MODE = ["VFO", "MEM", "CALL", "WX"]
TX_POWER = ["HIGH", "MED", "LOW"]
ANNOUNCE = ["OFF", "AUTO", "MANUAL"]
LANGUAGE = ["English", "Japanese"]
SQL_HANG_TIME = ["OFF", "125", "250", "500"]
MUTE_HANG_TIME = ["OFF", "125", "250", "500", "750", "1000"]
RECALL_METHOD = ["ALL", "CURRENT"]
SPEED = ["FAST", "SLOW"]
DTMF_PAUSE = ["100", "250", "500", "750", "1000", "1500", "2000"]
BACKLIGHT_COLOR = ["AMBER", "GREEN"]
KEY_FUNCTIONS = [
    "WX",
    "FREQBAND",
    "CTRL",
    "MONITOR",
    "VGS",
    "VOICE",
    "GROUP_UP",
    "MENU",
    "MUTE",
    "SHIFT",
    "DUAL",
    "MEM_TO_VFO",
    "VFO",
    "MR",
    "CALL",
    "MHZ",
    "TONE",
    "REV",
    "LOW",
    "LOCK",
    "A_B",
    "ENTER",
    "1750HZ",
]
SCAN_RESUME = ["TIME", "CARRIER", "SEEK"]
APO = ["OFF", "30", "60", "90", "120", "180"]
EXT_DATA_BAND = ["A", "B", "TXA_RXB", "TXB_RXA"]
EXT_DATA_SPEED = ["1200", "9600"]
SQC_SOURCE = ["OFF", "BUSY", "SQL", "TX", "BUSY_TX", "SQL_TX"]
STEP_SIZE = [5, 6.25, 28.33, 10, 12.5, 15, 20, 25, 30, 50, 100]

TONE_FREQUENCY = [
    67.0,
    69.3,
    71.9,
    74.4,
    77.0,
    79.7,
    82.5,
    85.4,
    88.5,
    91.5,
    94.8,
    97.4,
    100.0,
    103.5,
    107.2,
    110.9,
    114.8,
    118.8,
    123.0,
    127.3,
    131.8,
    136.5,
    141.3,
    146.2,
    151.4,
    156.7,
    162.2,
    167.9,
    173.8,
    179.9,
    186.2,
    192.8,
    203.5,
    240.7,
    210.7,
    218.1,
    225.7,
    229.1,
    233.6,
    241.8,
    250.3,
    254.1,
]

DCS_CODE = [
    23,
    25,
    26,
    31,
    32,
    36,
    43,
    47,
    51,
    53,
    54,
    65,
    71,
    72,
    73,
    74,
    114,
    115,
    116,
    122,
    125,
    131,
    132,
    134,
    143,
    145,
    152,
    155,
    156,
    162,
    165,
    172,
    174,
    205,
    212,
    223,
    225,
    226,
    243,
    244,
    245,
    246,
    251,
    252,
    255,
    261,
    263,
    265,
    266,
    271,
    274,
    306,
    311,
    315,
    325,
    331,
    332,
    343,
    346,
    351,
    356,
    364,
    365,
    371,
    411,
    412,
    413,
    423,
    431,
    432,
    445,
    446,
    452,
    454,
    455,
    462,
    464,
    465,
    466,
    503,
    506,
    516,
    523,
    565,
    532,
    546,
    565,
    606,
    612,
    624,
    627,
    631,
    632,
    654,
    662,
    664,
    703,
    712,
    723,
    731,
    732,
    734,
    743,
    754,
]

SHIFT_DIRECTION = ["SIMPLEX", "UP", "DOWN", "SPLIT"]

MODE = ["FM", "NFM", "AM"]


class SchemaDict(dict):
    def __init__(self, schema, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.schema = schema

    def items(self):
        return [(k, self[k]) for k in self.schema._declared_fields]


class RadioSchema(Schema):
    def from_tuple(self, values):
        """Read in values from a tuple

        Order of items in tuple must match order of declared fields
        for this schema."""

        data = self.load(dict(zip(self.declared_fields, values)))
        return data

    def to_tuple(self, obj, **kwargs):
        """Convert object to a tuple"""

        data = self.dump(obj, **kwargs)
        return [data[f] for f in self.declared_fields if f in data]

    def from_csv(self, s):
        """Read in values from a comma-separated string"""

        return self.from_tuple(s.split(","))

    def to_csv(self, obj, **kwargs):
        """Convert object to a comma-separated string"""

        return ",".join(self.to_tuple(obj, **kwargs))

    def to_raw_tuple(self, obj):
        return [obj[f] for f in self.declared_fields]

    def from_raw_tuple(self, values):
        return dict(zip(self.declared_fields, values))

    @post_load
    def add_schema(self, data, **kwargs):
        return SchemaDict(self, data)


class Indexed(Field):
    def __init__(self, values, fmt="{}", int_base=10, type=None, **kwargs):
        super().__init__(**kwargs)
        self.values = values
        self.fmt = fmt
        self.int_base = int_base
        self.type = type

    def _serialize(self, value, attr, obj, **kwargs):
        if self.type is not None:
            value = self.type(value)

        return self.fmt.format(self.values.index(value))

    def _deserialize(self, value, attr, data, **kwargs):
        res = self.values[int(value, self.int_base)]
        if self.type is not None:
            res = self.type(res)

        return res


class RadioFloat(Float):
    def __init__(self, length=10, **kwargs):
        super().__init__(**kwargs)
        self.length = length

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            value = 0

        fmt = "{:0" + str(self.length) + "d}"
        return fmt.format(int(float(value) * 1000000))

    def _deserialize(self, value, attr, data, **kwargs):
        return int(value) / 1000000.0


class RadioBoolean(Boolean):
    truthy = [1, "1", "true", "True", "TRUE", True]

    def _serialize(self, value, attr, obj, **kwargs):
        return "1" if value in self.truthy else "0"


class FormattedInteger(Integer):
    def __init__(self, fmt="{}", **kwargs):
        super().__init__(**kwargs)
        self.fmt = fmt

    def _serialize(self, value, attr, obj, **kwargs):
        return self.fmt.format(int(value))


class ME_Schema(RadioSchema):
    channel = FormattedInteger("{:03d}", required=True)
    rx_freq = RadioFloat(required=True)
    rx_step = Indexed(STEP_SIZE, required=True, type=float)
    shift = Indexed(SHIFT_DIRECTION, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True, type=float, default=67)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True, type=float, default=67)
    dcs_code = Indexed(
        values=DCS_CODE, required=True, fmt="{:03d}", type=int, default=23
    )
    offset = RadioFloat(length=8, required=True)
    mode = Indexed(MODE, required=True)
    tx_freq = RadioFloat(required=True)
    tx_step = Indexed(STEP_SIZE, required=True, type=float)
    lockout = RadioBoolean(required=True)
    name = String()

    export_fields = (
        "channel",
        "rx_freq",
        "rx_step",
        "shift",
        "reverse",
        "admit",
        "tone",
        "offset",
        "mode",
        "tx_freq",
        "tx_step",
        "lockout",
        "name",
    )

    @post_load
    def compute_admit_load(self, data, **kwargs):
        if data["tone_status"]:
            data["admit"] = "T"
            data["tone"] = data["tone_freq"]
        elif data["ctcss_status"]:
            data["admit"] = "C"
            data["tone"] = data["ctcss_freq"]
        elif data["dcs_status"]:
            data["admit"] = "D"
            data["tone"] = data["dcs_code"]
        else:
            data["admit"] = ""
            data["tone"] = ""

        return data

    @pre_dump
    def compute_admit_dump(self, obj, **kwargs):
        admit = obj["admit"]

        obj["tone_status"] = obj["ctcss_status"] = obj["dcs_status"] = False

        if admit == "T":
            obj["tone_status"] = True
            obj["tone_freq"] = obj["tone"]
        elif admit == "C":
            obj["ctcss_status"] = True
            obj["ctcss_freq"] = obj["tone"]
        elif admit == "D":
            obj["dcs_status"] = True
            obj["dcs_code"] = obj["tone"]

        return obj


class FO_Schema(RadioSchema):
    band = FormattedInteger(required=True, validate=Range(0, 1))
    rx_freq = RadioFloat(required=True)
    rx_step = Indexed(values=STEP_SIZE, required=True, type=float)
    shift = Indexed(SHIFT_DIRECTION, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True, type=float)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True, type=float)
    dcs_code = Indexed(values=DCS_CODE, required=True, fmt="{:03d}", type=int)
    offset = RadioFloat(length=8, required=True)
    mode = Indexed(MODE, required=True)


class CC_Schema(RadioSchema):
    index = FormattedInteger(required=True)
    rx_freq = RadioFloat(required=True)
    rx_step = Indexed(values=STEP_SIZE, required=True, type=float)
    shift = Indexed(SHIFT_DIRECTION, required=True)
    reverse = RadioBoolean(required=True)
    tone_status = RadioBoolean(required=True)
    ctcss_status = RadioBoolean(required=True)
    dcs_status = RadioBoolean(required=True)
    tone_freq = Indexed(values=TONE_FREQUENCY, required=True, fmt="{:02d}", type=float)
    ctcss_freq = Indexed(values=TONE_FREQUENCY, required=True, fmt="{:02d}", type=float)
    dcs_code = Indexed(values=DCS_CODE, required=True, fmt="{:03d}", type=int)
    offset = RadioFloat(length=8, required=True)
    mode = Indexed(MODE, required=True)
    tx_freq = RadioFloat(required=True)
    unknown = String(missing="0")


class TY_Schema(RadioSchema):
    model = String(validate=OneOf("K", "M"))
    mars_tx_expansion = Integer()
    max_tx_expansion = Integer()
    crossband = Integer()
    skycommand = Integer()


class BC_Schema(RadioSchema):
    ctrl = Integer()
    ptt = Integer()


class AE_Schema(RadioSchema):
    serial = String()
    extra = String()


class FV_Schema(RadioSchema):
    unit = Integer()
    v1 = String()
    v2 = String()
    v3 = String()
    v4 = String()


# 0,4,0,1,0,4,1,0,10,0,0,0,0,0,0,2,0,0,0,0,2,0,1,0,0,8,0,0,00,02,14,0D,0C,15,0,0,0,0,0,4,1,1
class MU_Schema(RadioSchema):
    beep = RadioBoolean(required=True)
    beep_volume = FormattedInteger(validate=validate.Range(min=0, max=7), required=True)
    external_speaker_mode = FormattedInteger(
        validate=validate.Range(min=0, max=2), required=True
    )
    announce = Indexed(ANNOUNCE, fmt="{:02X}", required=True)
    language = Indexed(LANGUAGE, fmt="{:02X}", required=True)
    voice_volume = FormattedInteger(
        validate=validate.Range(min=0, max=7), required=True
    )
    voice_speed = FormattedInteger(validate=validate.Range(min=0, max=4), required=True)
    playback_repeat = RadioBoolean(required=True)
    playback_repeat_interval = FormattedInteger(
        validate=validate.Range(min=0, max=60), required=True
    )
    continous_recording = RadioBoolean(required=True)
    vhf_aip = RadioBoolean(required=True)
    uhf_aip = RadioBoolean(required=True)
    s_meter_sql_hang_time = Indexed(SQL_HANG_TIME, fmt="{:02X}", required=True)
    mute_hang_time = Indexed(MUTE_HANG_TIME, fmt="{:02X}", required=True)
    beatshift = RadioBoolean(required=True)
    timeout_timer = FormattedInteger(
        validate=validate.Range(min=0, max=3), required=True
    )
    recall_method = Indexed(RECALL_METHOD, fmt="{:02X}", required=True)
    echolink_speed = Indexed(SPEED, fmt="{:02X}", required=True)
    dtmf_hold = RadioBoolean(required=True)
    dtmf_speed = Indexed(SPEED, fmt="{:02X}", required=True)
    dtmf_pause = Indexed(DTMF_PAUSE, fmt="{:02X}", required=True)
    dtmf_key_lock = RadioBoolean(required=True)
    auto_repeater_offset = RadioBoolean(required=True)
    hold_1750hz = RadioBoolean(required=True)
    unknown = String()
    brightness_level = FormattedInteger(required=True)
    auto_brightness = RadioBoolean(required=True)
    backlight_color = Indexed(BACKLIGHT_COLOR, fmt="{:02X}", required=True)
    pf1_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    pf2_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    mic_pf1_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    mic_pf2_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    mic_pf3_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    mic_pf4_key = Indexed(KEY_FUNCTIONS, fmt="{:02X}", required=True, int_base=16)
    mic_key_lock = RadioBoolean(required=True)
    scan_resume = Indexed(SCAN_RESUME, fmt="{:02X}", required=True)
    apo = Indexed(APO, fmt="{:02X}", required=True)
    ext_data_band = Indexed(EXT_DATA_BAND, fmt="{:02X}", required=True)
    ext_data_speed = Indexed(EXT_DATA_SPEED, fmt="{:02X}", required=True)
    sqc_source = Indexed(SQC_SOURCE, fmt="{:02X}", required=True)
    auto_pm_store = RadioBoolean(required=True)
    display_partition_bar = RadioBoolean(required=True)


ME_no_name = ME_Schema(exclude=["name"])

# Dynamically create instances of each schema

_classes = list(locals().values())
_globals = globals()
for cls in _classes:
    name = getattr(cls, "__name__", None)
    if name and name.endswith("_Schema"):
        varname = name[:-7]
        _globals[varname] = cls()



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
from kaitaistruct import __version__ as ks_version, KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(ks_version) < parse_version("0.7"):
    raise Exception(
        "Incompatible Kaitai Struct Python API: 0.7 or later is required, but you have %s"
        % (ks_version)
    )


class Memory(KaitaiStruct):
    """This is a Kaitai Struct[1] definition that parses the memory
    dump from a Kenwood TM-V71A radio. To generate a Python module from
    this description, use the Kaitai Struct Compiler:
    
        ksc --target python --outdir tmv71 memory.ksy
    
    This will generate `memory.py`.
    
    Example usage:
    
        from tmv71.memory import Memory
    
        data = Memory.from_file('my_dump_file.bin')
        for i, channel in enumerate(data.channels()):
          print('channel {} ({}) rx frequency: {}'.format(
                i, channel.name, channel.common.rx_freq))
    
    [1]: http://kaitai.io/
    """

    class DataBand(Enum):
        data_a = 0
        data_b = 1
        data_a_tx_b_rx = 2
        data_a_rx_b_tx = 3

    class ShiftDirection(Enum):
        simplex = 0
        up = 1
        down = 2
        invalid = 3

    class FreqBand(Enum):
        band_118 = 0
        band_144 = 1
        band_220 = 2
        band_300 = 3
        band_430 = 4
        band_1200 = 5

    class Modulation(Enum):
        fm = 0
        am = 1
        nfm = 2
        invalid = 255

    class PortSpeed(Enum):
        b9600 = 0
        b19200 = 1
        b38400 = 2
        b57600 = 3

    class ChannelBand(Enum):
        vhf = 5
        uhf = 8

    class DataSpeed(Enum):
        b1200 = 0
        b9600 = 1

    class TxPower(Enum):
        high = 0
        medium = 1
        low = 2

    class DisplayMode(Enum):
        vfo = 0
        memory = 1

    class Admit(Enum):
        none = 0
        dcs = 1
        ctcss = 2
        tone = 4
        invalid = 7

    class RepeaterIdtx(Enum):
        false = 0
        morse = 1
        voice = 2

    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.magic = self._io.read_bytes(2)

    class MiscSettings(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def wireless_remote(self):
            if hasattr(self, "_m_wireless_remote"):
                return (
                    self._m_wireless_remote
                    if hasattr(self, "_m_wireless_remote")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(17)
            self._m_wireless_remote = self._io.read_u1()
            self._io.seek(_pos)
            return (
                self._m_wireless_remote if hasattr(self, "_m_wireless_remote") else None
            )

        @property
        def repeater_hold(self):
            if hasattr(self, "_m_repeater_hold"):
                return (
                    self._m_repeater_hold if hasattr(self, "_m_repeater_hold") else None
                )

            _pos = self._io.pos()
            self._io.seek(30)
            self._m_repeater_hold = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_repeater_hold if hasattr(self, "_m_repeater_hold") else None

        @property
        def repeater_id(self):
            if hasattr(self, "_m_repeater_id"):
                return self._m_repeater_id if hasattr(self, "_m_repeater_id") else None

            _pos = self._io.pos()
            self._io.seek(368)
            self._m_repeater_id = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(6), 255, False)
            ).decode(u"ascii")
            self._io.seek(_pos)
            return self._m_repeater_id if hasattr(self, "_m_repeater_id") else None

        @property
        def pc_port_speed(self):
            if hasattr(self, "_m_pc_port_speed"):
                return (
                    self._m_pc_port_speed if hasattr(self, "_m_pc_port_speed") else None
                )

            _pos = self._io.pos()
            self._io.seek(33)
            self._m_pc_port_speed = self._root.PortSpeed(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_pc_port_speed if hasattr(self, "_m_pc_port_speed") else None

        @property
        def repeater_idtx(self):
            if hasattr(self, "_m_repeater_idtx"):
                return (
                    self._m_repeater_idtx if hasattr(self, "_m_repeater_idtx") else None
                )

            _pos = self._io.pos()
            self._io.seek(31)
            self._m_repeater_idtx = self._root.RepeaterIdtx(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_repeater_idtx if hasattr(self, "_m_repeater_idtx") else None

        @property
        def remote_id(self):
            if hasattr(self, "_m_remote_id"):
                return self._m_remote_id if hasattr(self, "_m_remote_id") else None

            _pos = self._io.pos()
            self._io.seek(18)
            self._m_remote_id = self._io.read_bytes(3)
            self._io.seek(_pos)
            return self._m_remote_id if hasattr(self, "_m_remote_id") else None

        @property
        def current_pm_channel(self):
            if hasattr(self, "_m_current_pm_channel"):
                return (
                    self._m_current_pm_channel
                    if hasattr(self, "_m_current_pm_channel")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(22)
            self._m_current_pm_channel = self._io.read_u1()
            self._io.seek(_pos)
            return (
                self._m_current_pm_channel
                if hasattr(self, "_m_current_pm_channel")
                else None
            )

        @property
        def key_lock(self):
            if hasattr(self, "_m_key_lock"):
                return self._m_key_lock if hasattr(self, "_m_key_lock") else None

            _pos = self._io.pos()
            self._io.seek(23)
            self._m_key_lock = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_key_lock if hasattr(self, "_m_key_lock") else None

        @property
        def crossband_repeat(self):
            if hasattr(self, "_m_crossband_repeat"):
                return (
                    self._m_crossband_repeat
                    if hasattr(self, "_m_crossband_repeat")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(16)
            self._m_crossband_repeat = self._io.read_u1()
            self._io.seek(_pos)
            return (
                self._m_crossband_repeat
                if hasattr(self, "_m_crossband_repeat")
                else None
            )

    class ExtendedFlagBits(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown = self._io.read_bits_int(7)
            self.lockout = self._io.read_bits_int(1) != 0

    class Channel(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            self.common = self._root.CommonVfoFields(self._io, self, self._root)

        @property
        def name(self):
            if hasattr(self, "_m_name"):
                return self._m_name if hasattr(self, "_m_name") else None

            self._m_name = self._root.channel_names[self.number]
            return self._m_name if hasattr(self, "_m_name") else None

        @property
        def extended_flags(self):
            if hasattr(self, "_m_extended_flags"):
                return (
                    self._m_extended_flags
                    if hasattr(self, "_m_extended_flags")
                    else None
                )

            self._m_extended_flags = self._root.channel_extended_flags[self.number]
            return (
                self._m_extended_flags if hasattr(self, "_m_extended_flags") else None
            )

    class CommonVfoFields(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.rx_freq_raw = self._io.read_u4le()
            self.rx_step_raw = self._io.read_u1()
            self.mod = self._root.Modulation(self._io.read_u1())
            self.flags = self._root.ChannelFlags(self._io, self, self._root)
            self.tone_frequency_raw = self._io.read_u1()
            self.ctcss_frequency_raw = self._io.read_u1()
            self.dcs_code_raw = self._io.read_u1()
            self.tx_offset_raw = self._io.read_u4le()
            self.tx_step_raw = self._io.read_u1()
            self.padding = self._io.read_u1()

        @property
        def ctcss_freq(self):
            if hasattr(self, "_m_ctcss_freq"):
                return self._m_ctcss_freq if hasattr(self, "_m_ctcss_freq") else None

            self._m_ctcss_freq = self._root.tables.tone_frequency[
                self.ctcss_frequency_raw
            ]
            return self._m_ctcss_freq if hasattr(self, "_m_ctcss_freq") else None

        @property
        def rx_freq(self):
            if hasattr(self, "_m_rx_freq"):
                return self._m_rx_freq if hasattr(self, "_m_rx_freq") else None

            self._m_rx_freq = self.rx_freq_raw / 1000000.0
            return self._m_rx_freq if hasattr(self, "_m_rx_freq") else None

        @property
        def dcs_code(self):
            if hasattr(self, "_m_dcs_code"):
                return self._m_dcs_code if hasattr(self, "_m_dcs_code") else None

            self._m_dcs_code = self._root.tables.dcs_code[self.dcs_code_raw]
            return self._m_dcs_code if hasattr(self, "_m_dcs_code") else None

        @property
        def tx_offset(self):
            if hasattr(self, "_m_tx_offset"):
                return self._m_tx_offset if hasattr(self, "_m_tx_offset") else None

            self._m_tx_offset = self.tx_offset_raw / 1000000.0
            return self._m_tx_offset if hasattr(self, "_m_tx_offset") else None

        @property
        def rx_step(self):
            if hasattr(self, "_m_rx_step"):
                return self._m_rx_step if hasattr(self, "_m_rx_step") else None

            self._m_rx_step = self._root.tables.step_size[self.rx_step_raw]
            return self._m_rx_step if hasattr(self, "_m_rx_step") else None

        @property
        def deleted(self):
            if hasattr(self, "_m_deleted"):
                return self._m_deleted if hasattr(self, "_m_deleted") else None

            if self.rx_freq_raw == 4294967295:
                self._m_deleted = True

            return self._m_deleted if hasattr(self, "_m_deleted") else None

        @property
        def tone_freq(self):
            if hasattr(self, "_m_tone_freq"):
                return self._m_tone_freq if hasattr(self, "_m_tone_freq") else None

            self._m_tone_freq = self._root.tables.tone_frequency[
                self.tone_frequency_raw
            ]
            return self._m_tone_freq if hasattr(self, "_m_tone_freq") else None

        @property
        def tx_step(self):
            if hasattr(self, "_m_tx_step"):
                return self._m_tx_step if hasattr(self, "_m_tx_step") else None

            self._m_tx_step = 0 if self.tx_step_raw == 255 else self.tx_step_raw
            return self._m_tx_step if hasattr(self, "_m_tx_step") else None

    class BandLimit(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.lower = self._io.read_u4le()
            self.upper = self._io.read_u4le()

    class ChannelExtendedFlags(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.band = self._io.read_u1()
            self.flags = self._root.ExtendedFlagBits(self._io, self, self._root)

    class Band(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            pass

        @property
        def freq_band_raw(self):
            if hasattr(self, "_m_freq_band_raw"):
                return (
                    self._m_freq_band_raw if hasattr(self, "_m_freq_band_raw") else None
                )

            _pos = self._io.pos()
            self._io.seek(2)
            self._m_freq_band_raw = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_freq_band_raw if hasattr(self, "_m_freq_band_raw") else None

        @property
        def freq_band(self):
            """For reasons known only to the designers, the constants that
            indicate frequency band are different on VFO A and VFO B. We
            can adjust the VFO B constants to match by substracing 4
            from the value.
            """
            if hasattr(self, "_m_freq_band"):
                return self._m_freq_band if hasattr(self, "_m_freq_band") else None

            self._m_freq_band = self._root.FreqBand(
                (self.freq_band_raw - (4 * self.number))
            )
            return self._m_freq_band if hasattr(self, "_m_freq_band") else None

        @property
        def display_mode(self):
            if hasattr(self, "_m_display_mode"):
                return (
                    self._m_display_mode if hasattr(self, "_m_display_mode") else None
                )

            _pos = self._io.pos()
            self._io.seek(1)
            self._m_display_mode = self._root.DisplayMode(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_display_mode if hasattr(self, "_m_display_mode") else None

        @property
        def tx_power(self):
            if hasattr(self, "_m_tx_power"):
                return self._m_tx_power if hasattr(self, "_m_tx_power") else None

            _pos = self._io.pos()
            self._io.seek(7)
            self._m_tx_power = self._root.TxPower(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_tx_power if hasattr(self, "_m_tx_power") else None

        @property
        def s_meter_squelch(self):
            if hasattr(self, "_m_s_meter_squelch"):
                return (
                    self._m_s_meter_squelch
                    if hasattr(self, "_m_s_meter_squelch")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(9)
            self._m_s_meter_squelch = self._io.read_u1()
            self._io.seek(_pos)
            return (
                self._m_s_meter_squelch if hasattr(self, "_m_s_meter_squelch") else None
            )

    class ProgramScanMemory(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.lower = self._root.CommonVfoFields(self._io, self, self._root)
            self.upper = self._root.CommonVfoFields(self._io, self, self._root)

    class BandMaskList(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            self.mask = [None] * (self.number)
            for i in range(self.number):
                self.mask[i] = self._io.read_u1()

    class VfoConfig(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.common = self._root.CommonVfoFields(self._io, self, self._root)

    class VfoSettingsList(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            self.list = [None] * (self.number)
            for i in range(self.number):
                self.list[i] = self._root.CommonVfoFields(self._io, self, self._root)

    class ProgramMemory(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            pass

        @property
        def band_limits(self):
            if hasattr(self, "_m_band_limits"):
                return self._m_band_limits if hasattr(self, "_m_band_limits") else None

            _pos = self._io.pos()
            self._io.seek(256)
            self._m_band_limits = [None] * (2)
            for i in range(2):
                self._m_band_limits[i] = self._root.BandLimitList(
                    5, self._io, self, self._root
                )

            self._io.seek(_pos)
            return self._m_band_limits if hasattr(self, "_m_band_limits") else None

        @property
        def beep(self):
            if hasattr(self, "_m_beep"):
                return self._m_beep if hasattr(self, "_m_beep") else None

            _pos = self._io.pos()
            self._io.seek(336)
            self._m_beep = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_beep if hasattr(self, "_m_beep") else None

        @property
        def data_speed(self):
            if hasattr(self, "_m_data_speed"):
                return self._m_data_speed if hasattr(self, "_m_data_speed") else None

            _pos = self._io.pos()
            self._io.seek(374)
            self._m_data_speed = self._root.DataSpeed(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_data_speed if hasattr(self, "_m_data_speed") else None

        @property
        def current_menu_item(self):
            if hasattr(self, "_m_current_menu_item"):
                return (
                    self._m_current_menu_item
                    if hasattr(self, "_m_current_menu_item")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(46)
            self._m_current_menu_item = self._io.read_u1()
            self._io.seek(_pos)
            return (
                self._m_current_menu_item
                if hasattr(self, "_m_current_menu_item")
                else None
            )

        @property
        def beep_volume(self):
            if hasattr(self, "_m_beep_volume"):
                return self._m_beep_volume if hasattr(self, "_m_beep_volume") else None

            _pos = self._io.pos()
            self._io.seek(337)
            self._m_beep_volume = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_beep_volume if hasattr(self, "_m_beep_volume") else None

        @property
        def ptt_band(self):
            if hasattr(self, "_m_ptt_band"):
                return self._m_ptt_band if hasattr(self, "_m_ptt_band") else None

            _pos = self._io.pos()
            self._io.seek(50)
            self._m_ptt_band = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_ptt_band if hasattr(self, "_m_ptt_band") else None

        @property
        def vfo_settings(self):
            if hasattr(self, "_m_vfo_settings"):
                return (
                    self._m_vfo_settings if hasattr(self, "_m_vfo_settings") else None
                )

            _pos = self._io.pos()
            self._io.seek(64)
            self._m_vfo_settings = [None] * (2)
            for i in range(2):
                self._m_vfo_settings[i] = self._root.VfoSettingsList(
                    5, self._io, self, self._root
                )

            self._io.seek(_pos)
            return self._m_vfo_settings if hasattr(self, "_m_vfo_settings") else None

        @property
        def group_link(self):
            if hasattr(self, "_m_group_link"):
                return self._m_group_link if hasattr(self, "_m_group_link") else None

            _pos = self._io.pos()
            self._io.seek(240)
            self._m_group_link = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(10), 255, False)
            ).decode(u"ascii")
            self._io.seek(_pos)
            return self._m_group_link if hasattr(self, "_m_group_link") else None

        @property
        def ctrl_band(self):
            if hasattr(self, "_m_ctrl_band"):
                return self._m_ctrl_band if hasattr(self, "_m_ctrl_band") else None

            _pos = self._io.pos()
            self._io.seek(51)
            self._m_ctrl_band = self._io.read_u1()
            self._io.seek(_pos)
            return self._m_ctrl_band if hasattr(self, "_m_ctrl_band") else None

        @property
        def name(self):
            if hasattr(self, "_m_name"):
                return self._m_name if hasattr(self, "_m_name") else None

            if self.number > 0:
                self._m_name = self._root.program_memory_names[(self.number - 1)]

            return self._m_name if hasattr(self, "_m_name") else None

        @property
        def band_masks(self):
            if hasattr(self, "_m_band_masks"):
                return self._m_band_masks if hasattr(self, "_m_band_masks") else None

            _pos = self._io.pos()
            self._io.seek(384)
            self._m_band_masks = [None] * (2)
            for i in range(2):
                self._m_band_masks[i] = self._root.BandMaskList(
                    5, self._io, self, self._root
                )

            self._io.seek(_pos)
            return self._m_band_masks if hasattr(self, "_m_band_masks") else None

        @property
        def bands(self):
            if hasattr(self, "_m_bands"):
                return self._m_bands if hasattr(self, "_m_bands") else None

            self._raw__m_bands = [None] * (2)
            self._m_bands = [None] * (2)
            for i in range(2):
                self._raw__m_bands[i] = self._io.read_bytes(12)
                io = KaitaiStream(BytesIO(self._raw__m_bands[i]))
                self._m_bands[i] = self._root.Band(i, io, self, self._root)

            return self._m_bands if hasattr(self, "_m_bands") else None

        @property
        def power_on_message(self):
            if hasattr(self, "_m_power_on_message"):
                return (
                    self._m_power_on_message
                    if hasattr(self, "_m_power_on_message")
                    else None
                )

            _pos = self._io.pos()
            self._io.seek(224)
            self._m_power_on_message = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(12), 0, False)
            ).decode(u"ascii")
            self._io.seek(_pos)
            return (
                self._m_power_on_message
                if hasattr(self, "_m_power_on_message")
                else None
            )

        @property
        def data_band(self):
            if hasattr(self, "_m_data_band"):
                return self._m_data_band if hasattr(self, "_m_data_band") else None

            _pos = self._io.pos()
            self._io.seek(373)
            self._m_data_band = self._root.DataBand(self._io.read_u1())
            self._io.seek(_pos)
            return self._m_data_band if hasattr(self, "_m_data_band") else None

    class ChannelFlags(KaitaiStruct):
        """These flags are included in byte 6 of the channel entry.
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.unknown = self._io.read_bits_int(1) != 0
            self.admit = self._root.Admit(self._io.read_bits_int(3))
            self.reverse = self._io.read_bits_int(1) != 0
            self.split = self._io.read_bits_int(1) != 0
            self.shift = self._root.ShiftDirection(self._io.read_bits_int(2))

    class Tables(KaitaiStruct):
        """This is a value-only collection of tables used to map
        numeric constants in the channel data into actual values.
        """

        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def step_size(self):
            if hasattr(self, "_m_step_size"):
                return self._m_step_size if hasattr(self, "_m_step_size") else None

            self._m_step_size = [5, 6.25, 28.33, 10, 12.5, 15, 20, 25, 30, 50, 100]
            return self._m_step_size if hasattr(self, "_m_step_size") else None

        @property
        def tone_frequency(self):
            if hasattr(self, "_m_tone_frequency"):
                return (
                    self._m_tone_frequency
                    if hasattr(self, "_m_tone_frequency")
                    else None
                )

            self._m_tone_frequency = [
                67,
                69.3,
                71.9,
                74.4,
                77,
                79.7,
                82.5,
                85.4,
                88.5,
                91.5,
                94.8,
                97.4,
                100,
                103.5,
                107.2,
                110.9,
                114.8,
                118.8,
                123,
                127.3,
                131.8,
                136.5,
                141.3,
                146.2,
                151.4,
                156.7,
                162.2,
                167.9,
                173.8,
                179.9,
                186.2,
                192.8,
                203.5,
                240.7,
                210.7,
                218.1,
                225.7,
                229.1,
                233.6,
                241.8,
                250.3,
                254.1,
            ]
            return (
                self._m_tone_frequency if hasattr(self, "_m_tone_frequency") else None
            )

        @property
        def dcs_code(self):
            if hasattr(self, "_m_dcs_code"):
                return self._m_dcs_code if hasattr(self, "_m_dcs_code") else None

            self._m_dcs_code = [
                23,
                25,
                26,
                31,
                32,
                36,
                43,
                47,
                51,
                53,
                54,
                65,
                71,
                72,
                73,
                74,
                114,
                115,
                116,
                122,
                125,
                131,
                132,
                134,
                143,
                145,
                152,
                155,
                156,
                162,
                165,
                172,
                174,
                205,
                212,
                223,
                225,
                226,
                243,
                244,
                245,
                246,
                251,
                252,
                255,
                261,
                263,
                265,
                266,
                271,
                274,
                306,
                311,
                315,
                325,
                331,
                332,
                343,
                346,
                351,
                356,
                364,
                365,
                371,
                411,
                412,
                413,
                423,
                431,
                432,
                445,
                446,
                452,
                454,
                455,
                462,
                464,
                465,
                466,
                503,
                506,
                516,
                523,
                565,
                532,
                546,
                565,
                606,
                612,
                624,
                627,
                631,
                632,
                654,
                662,
                664,
                703,
                712,
                723,
                731,
                732,
                734,
                743,
                754,
            ]
            return self._m_dcs_code if hasattr(self, "_m_dcs_code") else None

    class BandLimitList(KaitaiStruct):
        def __init__(self, number, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.number = number
            self._read()

        def _read(self):
            self.list = [None] * (self.number)
            for i in range(self.number):
                self.list[i] = self._root.BandLimit(self._io, self, self._root)

    @property
    def channels(self):
        if hasattr(self, "_m_channels"):
            return self._m_channels if hasattr(self, "_m_channels") else None

        _pos = self._io.pos()
        self._io.seek(5888)
        self._m_channels = [None] * (1000)
        for i in range(1000):
            self._m_channels[i] = self._root.Channel(i, self._io, self, self._root)

        self._io.seek(_pos)
        return self._m_channels if hasattr(self, "_m_channels") else None

    @property
    def program_memory(self):
        """The radio has 6 programmable memory regions. One is used when no
        PM channel is selected, and then there are 5 numbered PM channels.
        The PM channels store a variety of radio configuration settings so
        that you can quickly switch between them.
        """
        if hasattr(self, "_m_program_memory"):
            return (
                self._m_program_memory if hasattr(self, "_m_program_memory") else None
            )

        _pos = self._io.pos()
        self._io.seek(512)
        self._raw__m_program_memory = [None] * (6)
        self._m_program_memory = [None] * (6)
        for i in range(6):
            self._raw__m_program_memory[i] = self._io.read_bytes(512)
            io = KaitaiStream(BytesIO(self._raw__m_program_memory[i]))
            self._m_program_memory[i] = self._root.ProgramMemory(
                i, io, self, self._root
            )

        self._io.seek(_pos)
        return self._m_program_memory if hasattr(self, "_m_program_memory") else None

    @property
    def misc_settings(self):
        if hasattr(self, "_m_misc_settings"):
            return self._m_misc_settings if hasattr(self, "_m_misc_settings") else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._raw__m_misc_settings = self._io.read_bytes(512)
        io = KaitaiStream(BytesIO(self._raw__m_misc_settings))
        self._m_misc_settings = self._root.MiscSettings(io, self, self._root)
        self._io.seek(_pos)
        return self._m_misc_settings if hasattr(self, "_m_misc_settings") else None

    @property
    def echolink_names(self):
        if hasattr(self, "_m_echolink_names"):
            return (
                self._m_echolink_names if hasattr(self, "_m_echolink_names") else None
            )

        _pos = self._io.pos()
        self._io.seek(288)
        self._m_echolink_names = [None] * (10)
        for i in range(10):
            self._m_echolink_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(8), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_echolink_names if hasattr(self, "_m_echolink_names") else None

    @property
    def program_scan_memory(self):
        if hasattr(self, "_m_program_scan_memory"):
            return (
                self._m_program_scan_memory
                if hasattr(self, "_m_program_scan_memory")
                else None
            )

        _pos = self._io.pos()
        self._io.seek(21888)
        self._m_program_scan_memory = [None] * (10)
        for i in range(10):
            self._m_program_scan_memory[i] = self._root.ProgramScanMemory(
                self._io, self, self._root
            )

        self._io.seek(_pos)
        return (
            self._m_program_scan_memory
            if hasattr(self, "_m_program_scan_memory")
            else None
        )

    @property
    def group_names(self):
        if hasattr(self, "_m_group_names"):
            return self._m_group_names if hasattr(self, "_m_group_names") else None

        _pos = self._io.pos()
        self._io.seek(32000)
        self._m_group_names = [None] * (8)
        for i in range(8):
            self._m_group_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(16), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_group_names if hasattr(self, "_m_group_names") else None

    @property
    def wx_channel_names(self):
        if hasattr(self, "_m_wx_channel_names"):
            return (
                self._m_wx_channel_names
                if hasattr(self, "_m_wx_channel_names")
                else None
            )

        _pos = self._io.pos()
        self._io.seek(30688)
        self._m_wx_channel_names = [None] * (10)
        for i in range(10):
            self._m_wx_channel_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(8), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return (
            self._m_wx_channel_names if hasattr(self, "_m_wx_channel_names") else None
        )

    @property
    def dtmf_codes(self):
        if hasattr(self, "_m_dtmf_codes"):
            return self._m_dtmf_codes if hasattr(self, "_m_dtmf_codes") else None

        _pos = self._io.pos()
        self._io.seek(48)
        self._m_dtmf_codes = [None] * (10)
        for i in range(10):
            self._m_dtmf_codes[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(16), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_dtmf_codes if hasattr(self, "_m_dtmf_codes") else None

    @property
    def program_memory_names(self):
        if hasattr(self, "_m_program_memory_names"):
            return (
                self._m_program_memory_names
                if hasattr(self, "_m_program_memory_names")
                else None
            )

        _pos = self._io.pos()
        self._io.seek(32160)
        self._m_program_memory_names = [None] * (5)
        for i in range(5):
            self._m_program_memory_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(16), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return (
            self._m_program_memory_names
            if hasattr(self, "_m_program_memory_names")
            else None
        )

    @property
    def dtmf_names(self):
        if hasattr(self, "_m_dtmf_names"):
            return self._m_dtmf_names if hasattr(self, "_m_dtmf_names") else None

        _pos = self._io.pos()
        self._io.seek(208)
        self._m_dtmf_names = [None] * (10)
        for i in range(10):
            self._m_dtmf_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(8), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_dtmf_names if hasattr(self, "_m_dtmf_names") else None

    @property
    def channel_extended_flags(self):
        """The TM-V71A appears to maintain some extended channel flags. At
        the moment I have only identify the `lockout` flag, used to
        skip a channel during scans.
        """
        if hasattr(self, "_m_channel_extended_flags"):
            return (
                self._m_channel_extended_flags
                if hasattr(self, "_m_channel_extended_flags")
                else None
            )

        _pos = self._io.pos()
        self._io.seek(3584)
        self._m_channel_extended_flags = [None] * (1000)
        for i in range(1000):
            self._m_channel_extended_flags[i] = self._root.ChannelExtendedFlags(
                self._io, self, self._root
            )

        self._io.seek(_pos)
        return (
            self._m_channel_extended_flags
            if hasattr(self, "_m_channel_extended_flags")
            else None
        )

    @property
    def channel_names(self):
        if hasattr(self, "_m_channel_names"):
            return self._m_channel_names if hasattr(self, "_m_channel_names") else None

        _pos = self._io.pos()
        self._io.seek(22528)
        self._m_channel_names = [None] * (1000)
        for i in range(1000):
            self._m_channel_names[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(8), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_channel_names if hasattr(self, "_m_channel_names") else None

    @property
    def tables(self):
        """A list of lookup tables used to convert numeric constants stored
        in radio memory to their corresponding values.
        """
        if hasattr(self, "_m_tables"):
            return self._m_tables if hasattr(self, "_m_tables") else None

        self._m_tables = self._root.Tables(self._io, self, self._root)
        return self._m_tables if hasattr(self, "_m_tables") else None

    @property
    def echolink_codes(self):
        if hasattr(self, "_m_echolink_codes"):
            return (
                self._m_echolink_codes if hasattr(self, "_m_echolink_codes") else None
            )

        _pos = self._io.pos()
        self._io.seek(400)
        self._m_echolink_codes = [None] * (10)
        for i in range(10):
            self._m_echolink_codes[i] = (
                KaitaiStream.bytes_terminate(self._io.read_bytes(8), 255, False)
            ).decode(u"ascii")

        self._io.seek(_pos)
        return self._m_echolink_codes if hasattr(self, "_m_echolink_codes") else None



from contextlib import contextmanager
import csv
from functools import wraps
import hexdump
import logging
import serial
import struct
import sys
import time

from tmv71 import schema

LOG = logging.getLogger(__name__)
PORT_SPEED = ["9600", "19200", "38400", "57600"]
FREQUENCY_BAND = ["118", "144", "220", "300", "430", "1200"]
DTMF_TONES = "0123456789ABCD*#"

M_OFFSET_PORT_SPEED = 0x21
M_OFFSET_BANDA_BAND = 0x202
M_OFFSET_BANDB_BAND = 0x20E
M_OFFSET_OPERATING_MODE = 0x10
M_OFFSET_REMOTE_ID = 0x12


class RadioError(Exception):
    pass


class CommunicationError(RadioError):
    def __init__(self, msg=None):
        super().__init__(msg or self.__doc__)

    def __str__(self):
        return self.__doc__.format(*self.args)


class UnknownDeviceError(CommunicationError):
    """Unknown device: {}"""


class UnknownCommandError(CommunicationError):
    """The radio did not recognize the "{}" command"""


class InvalidCommandError(CommunicationError):
    """The radio was unable to execute the "{}" command"""


class UnexpectedResponseError(CommunicationError):
    """The radio returned an unexpected response"""


class ReadTimeoutError(CommunicationError):
    """A read operation timed out"""


class WrongModeError(RadioError):
    """Radio is not in programming mode"""


def schemacommand(schema):
    def decorator(f):
        @wraps(f)
        def inner(*args, **kwargs):
            return schema.from_tuple(f(*args, **kwargs))

        return inner

    return decorator


def bandcommand(f):
    """A decorator for commands that have (band, value) results"""

    @wraps(f)
    def _(*args, **kwargs):
        res = f(*args, **kwargs)
        return int(res[1])

    return _


def pm(f):
    """Decorator to ensure that radio is in programming mode"""

    @wraps(f)
    def _(self, *args, **kwargs):
        if not self._programming_mode:
            raise WrongModeError()

        return f(self, *args, **kwargs)

    if _.__doc__ is None:
        _.__doc__ = ""

    _.__doc__ += "\n\n[REQUIRES PROGRAMMING MODE]"

    return _


class TMV71:
    expected_id = "TM-V71"
    memory_max = 0x7F
    memory_magic = struct.pack("BB", 0x0, 0x4B)

    def __init__(self, port, speed=9600, debug=False, timeout=0.5):
        self.port = port
        self.speed = speed
        self.debug = debug
        self.timeout = timeout
        self._programming_mode = False
        self._ptt = False
        self.init_serial()

    def __repr__(self):
        return "<TMV71 on {0.port} @ {0.speed}>".format(self)

    def init_serial(self):
        LOG.info("opening %s at %d bps", self.port, self.speed)
        self._port = serial.Serial(
            self.port,
            baudrate=self.speed,
            rtscts=True,
            dsrdtr=True,
            timeout=self.timeout,
        )

    def write_bytes(self, data):
        """Write the given data to the radio."""

        if self.debug:
            print("write:", file=sys.stderr)
            print("\n".join(hexdump.dumpgen(data)), file=sys.stderr)
        self._port.write(data)

    def read_bytes(self, n=0, until=None):
        """Read n bytes of data from the radio."""

        if until is None:
            data = self._port.read(n)
        else:
            data = self._port.read_until(until)

        if not data:
            raise ReadTimeoutError()

        if self.debug:
            print("read:", file=sys.stderr)
            print("\n".join(hexdump.dumpgen(data)), file=sys.stderr)
        return data

    def read_line(self):
        """Read a carriage-return terminated line from the radio."""

        return self.read_bytes(until=b"\r")[:-1]

    def send_command_raw(self, command, *args):
        self.write_bytes(command)
        if args:
            args_encoded = b" " + b",".join(args)
            self.write_bytes(args_encoded)
        self.write_bytes(b"\r")
        return self.read_line()

    def send_command(self, *command):
        """Send a command to the radio.

        Convert the given command and arguments to bytes and send them to the
        radio.  Read a response, convert it into a string, and split it on
        commas.

            >>> radio.send_command('FV', 0)
            ['0', '1.00', '2.10', 'A', '1']

        All arguments are first converted to strings, so
        send_command('FV', 0) and send_command('FV', '0') are
        equivalent."""

        LOG.debug("sending command: %s", command)

        command_encoded = [str(arg).encode("ascii") for arg in command]
        res = self.send_command_raw(*command_encoded).decode("ascii")

        if res == "?":
            raise UnknownCommandError(command[0])
        elif res == "N":
            raise InvalidCommandError(command[0])
        elif not res.startswith(command[0]):
            raise UnexpectedResponseError(command[0])

        return res[3:].split(",")

    # ----------------------------------------------------------------------

    def reopen(self):
        """Close and re-open the serial port"""
        self._port.close()
        self._port.open()

    def clear(self):
        """Clear the communication channel.

        Ensure the radio is in a known state by attempting to exit
        programming mode, complete any partially entered command, and
        read any remaining output."""

        self.write_bytes(b"E\r")

        while True:
            try:
                self.read_bytes(1024)
            except ReadTimeoutError:
                break

        self.write_bytes(b"\r")
        res = self.read_line()
        if res != b"?":
            raise UnexpectedResponseError()

    def check_id(self):
        res = self.radio_id()
        LOG.debug("check_id: wanted %s, got %s", self.expected_id, res)
        if res != self.expected_id:
            raise UnknownDeviceError(res)

    def radio_id(self):
        """Return the radio ID"""
        return self.send_command("ID")[0]

    @schemacommand(schema.TY)
    def radio_type(self):
        """Return the radio type (K for the US, M for Europe)"""

        return self.send_command("TY")

    @schemacommand(schema.AE)
    def radio_serial(self):
        """Return the radio serial number"""

        return self.send_command("AE")

    @schemacommand(schema.FV)
    def radio_firmware(self):
        return self.send_command("FV", 0)

    def get_band_squelch(self, band):
        """Return the squelch setting for the given band"""

        return int(self.send_command("SQ", band)[0], 16)

    @bandcommand
    def get_band_squelch_state(self, band):
        """Return the squelch setting for the given band"""

        return self.send_command("BY", band)

    @bandcommand
    def get_band_reverse(self, band):
        """Return the state of reverse mode for the given band"""

        return self.send_command("AS", band)

    @bandcommand
    def set_band_reverse(self, band, reverse_state):
        """Return the state of reverse mode for the given band"""

        return self.send_command("AS", band, reverse_state)

    def get_lock_state(self):
        """Get the current state of the key lock"""
        return bool(self.send_command("LK")[0])

    def set_lock_state(self, lock_state):
        """Set the current state of the key lock"""
        return bool(self.send_command("LK", 1 if lock_state else 0)[0])

    def get_poweron_message(self):
        return self.send_command("MS")[0]

    def set_poweron_message(self, msg):
        return self.send_command("MS", msg)[0]

    def get_dual_band_mode(self):
        return int(self.send_command("DL")[0])

    def set_dual_band_mode(self):
        return int(self.send_command("DL", 0)[0])

    def set_single_band_mode(self):
        return int(self.send_command("DL", 1)[0])

    @bandcommand
    def get_channel(self, band):
        return self.send_command("MR", band)

    @bandcommand
    def set_channel(self, band, channel):
        channel = "{:03d}".format(channel)
        return self.send_command("MR", band, channel)

    @schemacommand(schema.BC)
    def get_ptt_ctrl(self):
        return self.send_command("BC")

    @schemacommand(schema.BC)
    def set_ptt_ctrl(self, ctrl, ptt):
        return self.send_command("BC", ctrl, ptt)

    def set_ptt(self, ptt_state):
        if ptt_state:
            self._ptt = True
            return self.send_command("TX")
        else:
            self._ptt = False
            return self.send_command("RX")

    @contextmanager
    def ptt(self):
        """A context manager that ensures ptt is released"""

        try:
            self.set_ptt(True)
            yield
        finally:
            self.set_ptt(False)

    dtmf_tone_slow = 0.250
    dtmf_tone_fast = 0.040

    def send_dtmf(self, tones, fast=False):
        """Send a series of DTMF tones.

        This will send each tone for <delay> seconds (default 0.250).
        When sending tones there are no "spaces" between tones. This
        appears to be a limitation of the underlying `DT` command.
        """

        delay = self.dtmf_tone_fast if fast else self.dtmf_tone_slow

        time.sleep(delay)
        for tone in tones:
            self.send_command("DT", 0, "{:X}".format(DTMF_TONES.index(tone.upper())))
            time.sleep(delay)

    @bandcommand
    def get_band_mode(self, band):
        return self.send_command("VM", band)

    @bandcommand
    def set_band_mode(self, band, mode):
        return self.send_command("VM", band, mode)

    @bandcommand
    def get_tx_power(self, band):
        return self.send_command("PC", band)

    @bandcommand
    def set_tx_power(self, band, power):
        return self.send_command("PC", band, power)

    @schemacommand(schema.FO)
    def get_band_vfo(self, band):
        return self.send_command("FO", band)

    @schemacommand(schema.FO)
    def set_band_vfo(self, band, settings):
        settings["band"] = band
        return self.send_command("FO", schema.FO.to_csv(settings))

    @schemacommand(schema.CC)
    def get_call_channel(self, index):
        return self.send_command("CC", index)

    @schemacommand(schema.CC)
    def set_call_channel(self, index, settings):
        settings[index] = index
        return self.send_command("CC", schema.CC.to_csv(settings))

    @schemacommand(schema.ME)
    def get_channel_entry(self, channel):
        res = self.send_command("ME", "{:03d}".format(channel))
        res.append(self.get_channel_name(channel))
        return res

    def set_channel_entry(self, channel, settings):
        settings[channel] = channel
        self.send_command("ME", schema.ME_no_name.to_csv(settings))
        self.set_channel_name(channel, settings["name"])

    def delete_channel_entry(self, channel):
        channel = "{:03d}".format(channel)
        try:
            return self.send_command("ME", channel, "")
        except InvalidCommandError:
            return [""]

    def get_channel_name(self, channel):
        channel = "{:03d}".format(int(channel))
        res = self.send_command("MN", channel)
        return res[1]

    def set_channel_name(self, channel, name):
        channel = "{:03d}".format(int(channel))
        res = self.send_command("MN", channel, name)
        return res[1]

    @schemacommand(schema.MU)
    def get_radio_config(self):
        return self.send_command("MU")

    @schemacommand(schema.MU)
    def set_radio_config(self, settings):
        return self.send_command("MU", schema.MU.to_csv(settings))

    # ----------------------------------------------------------------------

    @pm
    def get_port_speed(self):
        speed = self.read_block(M_OFFSET_PORT_SPEED, 1)
        speed = struct.unpack("B", speed)[0]
        return PORT_SPEED[speed]

    @pm
    def set_port_speed(self, speed):
        speed = PORT_SPEED.index(speed)
        self.write_block(M_OFFSET_PORT_SPEED, bytes([speed]))

    @pm
    def get_frequency_band(self, band):
        if band == 0:
            address = M_OFFSET_BANDA_BAND
        elif band == 1:
            address = M_OFFSET_BANDB_BAND
        else:
            raise ValueError("invalid band: {}".format(band))

        res = self.read_block(address, 1)
        res = struct.unpack("B", res)[0]

        # The constants in FREQUENCY_BAND are correct for band A, but
        # for band B we need to add 4 to the value.
        res -= 4 * band

        return FREQUENCY_BAND[res]

    @pm
    def set_frequency_band(self, band, freq_band):
        freq_band = FREQUENCY_BAND.index(freq_band)

        # The constants in FREQUENCY_BAND are correct for band A, but
        # for band B we need to add 4 to the value.
        freq_band += 4 * band

        if band == 0:
            address = M_OFFSET_BANDA_BAND
        elif band == 1:
            address = M_OFFSET_BANDB_BAND
        else:
            raise ValueError("invalid band: {}".format(band))

        self.write_block(address, bytes([freq_band]))

    @pm
    def reset(self):
        """Reset to default configuration"""
        self.write_block(0, b"\xff")

    @pm
    def get_operating_mode(self):
        """Get current radio operating mode"""

        res = self.read_block(M_OFFSET_OPERATING_MODE, 2)
        return struct.unpack("BB", res)

    @pm
    def set_operating_mode(self, repeater, wireless):
        wireless = 1 if wireless else 0
        repeater = 1 if repeater else 0

        self.write_block(M_OFFSET_OPERATING_MODE, struct.pack("BB", repeater, wireless))

    @pm
    def get_remote_id(self):
        res = self.read_block(0x12, 3)
        return res

    @pm
    def set_remote_id(self, remote_id):
        if len(remote_id) != 3:
            raise ValueError("remote id must be three digits")

        self.write_block(0x12, remote_id)

    # ----------------------------------------------------------------------

    @contextmanager
    def programming_mode(self):
        """Wrap code that interacts with the radio in programming mode.

        You must enclose calls that require programming mode with this
        context manager.  For example:

            with radio.programming_mode():
                radio.read_block(0, 0)

        Failure to use the context manager will result in
        WrongModeError exception."""

        self.enter_programming_mode()
        try:
            yield
        finally:
            self.exit_programming_mode()

    def enter_programming_mode(self):
        LOG.debug("entering programming mode")
        self.write_bytes(b"0M PROGRAM\r")
        res = self.read_line()
        if res != b"0M":
            raise UnexpectedResponseError()
        self._programming_mode = True

    def exit_programming_mode(self):
        LOG.debug("exiting programming mode")
        self.write_bytes(b"E")
        self._programming_mode = False
        for expected in [b"\x06", b"\r", b"\x00"]:
            res = self.read_bytes(1)
            if res != expected:
                raise UnexpectedResponseError()

    @pm
    def read_block(self, address, size):
        """Read data from the radio"""

        LOG.debug("read address %d, size %d", address, size)

        args = struct.pack(">HB", address, size)

        self.write_bytes(b"R" + args)
        res = self.read_bytes(4)
        if res != b"W" + args:
            raise UnexpectedResponseError(res)

        data = self.read_bytes(size if size else 256)
        self.write_bytes(bytes([6]))
        self.check_ack()
        return data

    @pm
    def write_block(self, address, data):
        """Write data to the radio"""

        size = len(data)
        if size > 256:
            raise ValueError("write_block cannot write more than 256 bytes")

        LOG.debug("write address %d, size %d", address, size)
        if size == 256:
            size = 0

        self.write_bytes(struct.pack(">BHB", ord("W"), address, size))
        self.write_bytes(bytes(data))
        self.check_ack()

    def check_ack(self):
        """Validate the response to programming mode commands."""

        res = self.read_bytes(1)
        if res == b"\x15":
            LOG.warning("radio is in error state (continuing)")
        elif res != b"\x06":
            raise UnexpectedResponseError()

    @pm
    def memory_dump(self, fd):
        """Read data from the radio and write it to a file-like object."""

        for block in range(self.memory_max):
            addr = block * 256
            LOG.debug("reading block %d", block)
            data = self.read_block(addr, 0)
            fd.write(data)

    @pm
    def memory_restore(self, fd, force=False):
        """Read data from a file-like object and write it to the radio.

        This command is similar to the behavior of the MCP-2A "Write
        data to the receiver" command"; in particular, it initially
        writes 0xFF to address 0 before proceeding to load the data
        from the input file descriptor to the radio. This will cause
        the radio to reset to defaults if the write operation is
        interrupted."""

        magiclen = len(self.memory_magic)

        # Check that radio state seems sane
        data = self.read_block(0, magiclen)
        if data != self.memory_magic:
            if force:
                LOG.warning("Radio does not contain expected " "data (continuing)")
            else:
                raise ValueError("radio does not contain expected data")

        # Check that input data seems sane
        data = fd.read(magiclen)
        if data != self.memory_magic:
            if force:
                LOG.warning("Input does not contain expected " "data (continuing)")
            else:
                raise ValueError("input does not contain expected data")

        self.write_block(0, b"\xff")

        fd.seek(magiclen)
        self.write_block(magiclen, fd.read(256 - magiclen))

        for block in range(1, self.memory_max):
            addr = block * 256
            LOG.debug("writing block %d", block)
            data = fd.read(256)
            if not data:
                raise EOFError("Ran out of data in memory_restore")
            self.write_block(addr, data)

        self.write_block(0, self.memory_magic)

    # ----------------------------------------------------------------------

    def import_channels(self, fd, selected=None, ignore_errors=False, sync=False):
        """Import channels from a CSV document.

        - fd: A file-like object
        - selected: A list of channels to import. If this is None
          (or empty), try to import all channels.
        - ignore_errors: if a channel cannot be imported, continue
          trying to import subsequent channels
        - sync: delete channels in the radio that are not present
          in the input.
        """

        selected = selected if selected else range(1000)
        fields = schema.ME.export_fields

        reader = csv.DictReader(fd, fields)
        channelmap = {}
        for row in reader:
            # skip the header row if we find one
            if row["channel"] == "channel":
                continue

            # a deleted channel has no rx_freq
            if not row["rx_freq"]:
                continue

            channelmap[int(row["channel"])] = row

        for channel in selected:
            if channel not in channelmap:
                if sync:
                    LOG.info("deleting channel %d", channel)
                    self.delete_channel_entry(channel)
            else:
                LOG.info("setting information for channel %d", channel)
                try:
                    self.set_channel_entry(channel, channelmap[channel])
                except InvalidCommandError:
                    if ignore_errors:
                        LOG.warning("Unable to set channel %d", channel)
                    else:
                        raise

    def export_channels(self, fd, selected=None, skip_deleted=False):
        """Export channels to a CSV document

        - fd: A file-like object
        - selected: A list of channels to export. If this is None
          (or empty), try to export all channels.
        - skip_deleted: do not emit entries for deleted channels.
        """

        selected = selected if selected else range(1000)
        fields = schema.ME.export_fields

        writer = csv.DictWriter(fd, fields)
        writer.writeheader()
        for channel in selected:
            LOG.info("getting information for channel %d", channel)

            try:
                channel_config = self.get_channel_entry(channel)

                # csv.DictWriter complains about unknown fields
                channel_config = {k: channel_config[k] for k in fields}

                writer.writerow(channel_config)
            except InvalidCommandError:
                LOG.debug("channel %d does not exist", channel)
                if not skip_deleted:
                    writer.writerow({"channel": channel})


class TMD710(TMV71):
    expected_id = "TM-D710"
    memory_magic = struct.pack("BB", 0x0, 0x4D)
    memory_max = 0xFF



import click
import enum
import functools
import hexdump
import json
import logging
import os
import shlex
import sys
import tabulate
import time

from tmv71 import __version__
from tmv71 import api
from tmv71 import schema

TMV71_CONFIG = os.path.expanduser(
    os.path.join(os.environ.get("XDG_CONFIG_HOME", "~/.config"), "tmv71.json")
)

LOG = logging.getLogger(__name__)
BAND_NAMES = ["A", "B", "0", "1"]


class FORMATS(enum.Enum):
    SHELL = 0
    TABLE = 1
    JSON = 2


def fmt_dict(d, format=FORMATS.SHELL, table_format="grid", key=None):
    """Format a dictionary for output"""

    if isinstance(format, str):
        format = getattr(FORMATS, format.upper())

    if key:
        d = {k: d[k] for k in key}

    if format == FORMATS.SHELL:
        return "\n".join(
            "{}={}".format(k, shlex.quote("{}".format(v))) for k, v in d.items()
        )
    elif format == FORMATS.JSON:
        return json.dumps(d, indent=2)
    elif format == FORMATS.TABLE:
        return tabulate.tabulate(d.items(), tablefmt=table_format)


def formatted(f):
    """Decoration for commands that produce formatted output.

    Adds formatting options to the command, takes a dictionary
    returned from the command and passes it to fmt_dict for output."""

    @click.option(
        "--format",
        "-F",
        type=click.Choice([x.name.lower() for x in FORMATS]),
        default=list(FORMATS)[0].name.lower(),
    )
    @click.option(
        "--table-format",
        "-T",
        default="grid",
        type=click.Choice(tabulate.tabulate_formats),
    )
    @click.option(
        "--key",
        "-K",
        multiple=True,
        help="Limit output to the specified key (may be " "specified multiple times)",
    )
    @functools.wraps(f)
    def _(format, table_format, key, *args, **kwargs):
        res = f(*args, **kwargs)
        print(fmt_dict(res, format, table_format, key))

    return _


def apply_options_from_schema(model, **kwargs):
    """Create command line options for fields described in a schema"""

    options = []

    for fname, fspec in model.declared_fields.items():
        display_name = fname.replace("_", "-")
        if fname == "band":
            pass
        elif isinstance(fspec, schema.RadioBoolean):
            options.append(
                click.option(
                    "--{0}/--no-{0}".format(display_name),
                    default=None,
                    **kwargs.get(fname, {})
                )
            )
        elif isinstance(fspec, schema.Indexed):
            values = [str(v) for v in fspec.values]
            options.append(
                click.option(
                    "--{}".format(display_name),
                    type=click.Choice(values),
                    **kwargs.get(fname, {})
                )
            )
        elif isinstance(fspec, schema.FormattedInteger):
            options.append(
                click.option(
                    "--{}".format(display_name), type=int, **kwargs.get(fname, {})
                )
            )
        elif isinstance(fspec, schema.RadioFloat):
            options.append(
                click.option(
                    "--{}".format(display_name), type=float, **kwargs.get(fname, {})
                )
            )

    def _apply_options(f):
        for option in reversed(options):
            f = option(f)

        return f

    return _apply_options


def clear_first(f):
    """Clear the communication channel.

    This decorator will use the TMV71.clear() method to attempt to put the
    radio in a known state before running a command.  The top-level option
    --no-clear (-K) will skip this step (or you can set TMV71_NO_CLEAR=1 in
    your environment), and --clear-retries (-R, or TMV71_CLEAR_RETRIES)
    controls how many times it will retry before failing with an error."""

    @functools.wraps(f)
    def _(ctx, *args, **kwargs):
        if not ctx.settings.no_clear:
            LOG.info("clearing communication channel")
            for i in range(ctx.settings.clear_retries + 1):
                try:
                    ctx.api.clear()
                except api.CommunicationError:
                    LOG.info("no response from the radio (try %d)", i)
                    ctx.api.reopen()
                else:
                    break
            else:
                raise click.ClickException("failed to communicate with the radio")
        return f(ctx, *args, **kwargs)

    return _


class ApplicationSettings:
    no_clear = False
    clear_retries = 0
    port = "/dev/ttyS0"
    speed = 9600
    verbose = 0


SETTINGS = ApplicationSettings()


class ApplicationContext:
    """Store settings and global state for the cli"""

    def __init__(self, settings):
        self.settings = settings
        self.init_api()

    def init_api(self):
        self.api = api.TMV71(
            port=self.settings.port,
            speed=self.settings.speed,
            debug=(self.settings.verbose > 2),
        )


def safe_main():
    """Wrap commands to catch and report expected exceptions"""

    res = 1

    try:
        main()
    except Exception as e:
        if SETTINGS.verbose > 1:
            raise
        else:
            LOG.error("%s", e)
    else:
        res = 0

    sys.exit(res)


@click.group(context_settings=dict(auto_envvar_prefix="TMV71"))
@click.option(
    "-f",
    "--config-file",
    default=TMV71_CONFIG,
    show_default=True,
    help="Path to a JSON configuration file",
)
@click.option(
    "--no-config", is_flag=True, help="Do not attempt to read any configuration file"
)
@click.option("-p", "--port", help="Path to a serial device")
@click.option(
    "-s", "--speed", type=click.Choice(api.PORT_SPEED), help="Set the port speed"
)
@click.option(
    "-K",
    "--no-clear",
    is_flag=True,
    default=None,
    help="Skip initial clear operation before running command",
)
@click.option(
    "-R",
    "--clear-retries",
    type=int,
    help="Number of times to retry clear operation " "before failing",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Increase verbosity. May be specified multiple times.",
)
@click.pass_context
def main(ctx, config_file, no_config, **kwargs):
    global SETTINGS

    # Override default settings with settings from configuration file
    if not no_config:
        try:
            with open(config_file) as fd:
                conf = json.load(fd)
            LOG.info("loaded configuration from %s", config_file)
            for k, v in conf.items():
                setattr(SETTINGS, k, v)
        except IOError:
            pass

    # Override settings from configuration file with settings from
    # command line
    for k, v in kwargs.items():
        if v is not None:
            setattr(SETTINGS, k, v)

    try:
        loglevel = ["WARNING", "INFO", "DEBUG"][SETTINGS.verbose]
    except IndexError:
        loglevel = "DEBUG"

    logging.basicConfig(level=loglevel)

    ctx.obj = ApplicationContext(SETTINGS)


@main.command()
def version():
    """Print the software version"""
    print("{}".format(__version__))


@main.command()
@click.argument("message", default=None, required=False)
@click.pass_obj
@clear_first
def poweron_message(ctx, message):
    """Get or set the power on message."""

    if message is None:
        res = ctx.api.get_poweron_message()
    else:
        res = ctx.api.set_poweron_message(message)

    print(res[0])


def normalize_band(band):
    """Turns a band name into a numeric 0 or 1"""

    if band in ["0", "A"]:
        return 0
    elif band in ["1", "B"]:
        return 1
    else:
        raise ValueError("invalid band number: {}".format(band))


@main.command()
@click.option("--on", "ptt_state", flag_value=True)
@click.option("--off", "ptt_state", flag_value=False, default=False)
@click.pass_obj
@clear_first
def ptt(ctx, ptt_state):
    """Activate or deactivate transmitter."""

    ctx.api.set_ptt(ptt_state)


@main.command()
@click.option("--lock/--unlock", "lock_state", default=None)
@click.pass_obj
@clear_first
def lock(ctx, lock_state):
    """Lock or unlock the key lock"""

    if lock_state is None:
        res = ctx.api.get_lock_state()
    else:
        res = ctx.api.set_lock_state(lock_state)

    print(*res)


@main.command()
@click.option("-w", "--wait", is_flag=True)
@click.option("--slow/--fast", "fast")
@click.argument("tones", nargs=-1)
@click.pass_obj
@clear_first
def send_dtmf(ctx, wait, fast, tones):
    """Send DTMF tones"""

    tones = "".join(tones).replace(" ", "")
    if wait:
        LOG.info("waiting until channel is free")
        band_state = ctx.api.get_ptt_ctrl()
        ptt = band_state["ptt"]
        LOG.debug("current ptt band is %s", ptt)
        while ctx.api.get_band_squelch_state(ptt) == 1:
            LOG.debug("waiting until band %s is not busy", ptt)
            time.sleep(0.5)

    with ctx.api.ptt():
        ctx.api.send_dtmf(tones, fast)


# ----------------------------------------------------------------------


@main.group()
def band():
    """Commands for controlling the dual bands"""
    pass


@band.command("squelch")
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.pass_obj
@clear_first
def band_squelch(ctx, band):
    """Get current squelch setting for the specified band."""

    band = normalize_band(band)
    res = ctx.api.get_band_squelch(band)
    print(res)


@band.command("squelch-state")
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.pass_obj
@clear_first
def band_squelch_state(ctx, band):
    """Report whether squelch is open on the specified band"""

    band = normalize_band(band)
    res = ctx.api.get_band_squelch_state(band)
    print(res)


@band.command("reverse")
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.option("--on/--off", "reverse", default=None)
@click.pass_obj
@clear_first
def band_reverse(ctx, band, reverse):
    """Activate reverse on the specified band"""

    band = normalize_band(band)
    if reverse is None:
        res = ctx.api.get_band_reverse(band)
    else:
        reverse = 1 if reverse else 0
        res = ctx.api.set_band_reverse(band, reverse)

    print(res)


@band.command("mode")
@click.option("--vfo", "mode", flag_value=schema.BAND_MODE.index("VFO"))
@click.option("--mem", "--memory", "mode", flag_value=schema.BAND_MODE.index("MEM"))
@click.option("--call", "mode", flag_value=schema.BAND_MODE.index("CALL"))
@click.option("--wx", "--weather", "mode", flag_value=schema.BAND_MODE.index("WX"))
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.pass_obj
@clear_first
def band_mode(ctx, mode, band):
    """Set selected band to VFO, MEM, call channel, or weather."""

    band = normalize_band(band)
    if mode is None:
        res = ctx.api.get_band_mode(band)
    else:
        res = ctx.api.set_band_mode(band, int(mode))

    print(schema.BAND_MODE[res])


@band.command("select")
@click.option("-1", "--single", "mode", flag_value=1)
@click.option("-2", "--dual", "mode", flag_value=2)
@click.option("-c", "--control", "which", flag_value="control")
@click.option("-p", "--ptt", "which", flag_value="ptt")
@click.argument("band", required=False)
@formatted
@click.pass_obj
@clear_first
def band_select(ctx, band, which, mode):
    """Select control and ptt band, and select single or dual-band mode

    Use band A as control band and band B as ptt band:

        tmv71 band select A --control
        tmv71 band select B --ptt

    Use band B for both control and ptt:

        tmv71 band select B

    Use band B in single band mode:

        tmv71 band select B -1

    Place radio back into dual-band mode:

        tmv71 band select -2
    """

    if which and mode:
        raise click.ClickException("-1/-2 and -c/-p are mutually exclusive")

    cur_state = ctx.api.get_ptt_ctrl()
    cur_mode = ctx.api.get_dual_band_mode()

    if mode == 2:
        cur_mode = ctx.api.set_dual_band_mode()
    elif band is not None:
        band = normalize_band(band)

        if mode:
            if mode == 1:
                cur_state["ctrl"] = cur_state["ptt"] = band
                cur_mode = ctx.api.set_single_band_mode()
        else:
            if which == "control":
                cur_state["ctrl"] = band
            elif which == "ptt":
                cur_state["ptt"] = band
            else:
                cur_state["ctrl"] = cur_state["ptt"] = band

        cur_state = ctx.api.set_ptt_ctrl(**cur_state)

    return dict(mode=["dual", "single"][cur_mode], **cur_state)


@band.command()
@click.option("--low", "power", flag_value=schema.TX_POWER.index("LOW"))
@click.option("--medium", "--med", "power", flag_value=schema.TX_POWER.index("MED"))
@click.option("--high", "power", flag_value=schema.TX_POWER.index("HIGH"))
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.pass_obj
@clear_first
def txpower(ctx, power, band):
    """Set tx power for the selected band."""

    band = normalize_band(band)
    if power is None:
        res = ctx.api.get_tx_power(band)
    else:
        res = ctx.api.set_tx_power(band, int(power))

    print(schema.TX_POWER[res])


def resolve_range(rspec):
    """Resolves a range specification into a list of numbers.

    A range specification is a list of numbers and ranges, where a range
    is specified as lower_bound:upper_bound. Given a range specification
    of the form:

        [1, 3, 5, 7:10, 18:20]

    The return value would be:

        [1, 3, 5, 7, 8, 9, 10, 18, 19, 20]

    This method is used by the channel import and export commands.
    """

    selected = []

    if rspec:
        for entry in rspec:
            if ":" in entry:
                r_start, r_end = (int(x) for x in entry.split(":"))
                selected.extend(range(r_start, r_end + 1))
            else:
                selected.append(int(entry))

    return selected


@main.command()
@click.argument("speed", type=click.Choice(api.PORT_SPEED), required=False)
@click.pass_obj
@clear_first
def port_speed(ctx, speed):
    """Get or set the PC port speed.

    Note that because this command involves reading from/writing to
    memory directly, it will briefly reset the radio.

    Valid port speeds are 9600, 19200, 38400, and 57600."""

    with ctx.api.programming_mode():
        if speed is None:
            LOG.info("getting port speed")
            speed = ctx.api.get_port_speed()
        else:
            LOG.info("setting port speed to %d bps", speed)
            ctx.api.set_port_speed(speed)

    print(speed)


@main.command("set")
@apply_options_from_schema(schema.MU)
@click.option(
    "--reset",
    is_flag=True,
    help="Some options (e.g. brightness-level) require "
    "a reset before they will take effect",
)
@formatted
@click.pass_obj
@clear_first
def radio_set(ctx, reset, **kwargs):
    """Get or set various configuration options."""
    res = ctx.api.get_radio_config()

    set_radio = False
    for k, v in kwargs.items():
        if v is None:
            continue

        set_radio = True
        res[k] = v

    if set_radio:
        LOG.info("configuring radio")
        res = ctx.api.set_radio_config(res)

    if reset:
        with ctx.api.programming_mode():
            pass

    return res


@main.command()
@click.option("--wireless/--no-wireless", default=None)
@click.option("--repeater/--no-repeater", default=None)
@click.option("--normal", is_flag=True, default=None)
@click.pass_obj
@clear_first
def op_mode(ctx, wireless, repeater, normal):
    """Control the radio operating mode.

    Use this to activate the cross-band repeater feature (--repeater),
    the wireless remote control (--wireless), or both. Use --normal to
    return to normal operation.
    """
    with ctx.api.programming_mode():
        for opt in [wireless, repeater, normal]:
            if opt is not None:
                break
        else:
            print(*ctx.api.get_operating_mode())
            return

        if normal:
            repeater = 0
            wireless = 0

        ctx.api.set_operating_mode(repeater, wireless)


@main.command()
@click.argument("remote_id", required=False)
@click.pass_obj
@clear_first
def remote_id(ctx, remote_id):
    """Get or set the remote id.

    This is the access code require to control the radio in wireless
    remote mode."""

    with ctx.api.programming_mode():
        if remote_id is None:
            res = ctx.api.get_remote_id()
            print(res.decode("ascii"))
        else:
            ctx.api.set_remote_id(remote_id.encode("ascii"))


# ----------------------------------------------------------------------


@main.group()
def channel():
    """Commands for interacting with memory channels"""
    pass


@channel.command("tune")
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.argument("channel", type=int, required=False)
@click.pass_obj
@clear_first
def channel_tune(ctx, band, channel):
    """Get or set the memory channel for the selected band."""

    band = normalize_band(band)
    if channel is None:
        res = ctx.api.get_channel(band)
    else:
        res = ctx.api.set_channel(band, channel)

    print("{:03d}".format(res))


@channel.command()
@apply_options_from_schema(schema.ME)
@click.option("-n", "--name")
@click.argument("channel", type=int)
@formatted
@click.pass_obj
@clear_first
def entry(ctx, channel, name, **kwargs):
    """View or edit memory channels."""

    res = ctx.api.get_channel_entry(channel)

    set_radio = False
    for k, v in kwargs.items():
        if v is None:
            continue

        set_radio = True
        res[k] = v

    if name is not None:
        LOG.info("setting name for channel %s", channel)
        ctx.api.set_channel_name(channel, name)

    if set_radio:
        LOG.info("configuring channel %s", channel)
        res = ctx.api.set_channel_entry(channel, res)
        res = ctx.api.get_channel_entry(channel)

    return res


@channel.command("export")
@click.option("-o", "--output", type=click.File("w"), default=sys.stdout)
@click.option(
    "-c",
    "--channels",
    multiple=True,
    help="Specify a single chanel (-c 1) or " "a range of channels (-c 1:10)",
)
@click.option(
    "-s", "--skip-deleted", is_flag=True, help="Do not export deleted channels"
)
@click.pass_obj
@clear_first
def export_channels(ctx, output, channels, skip_deleted):
    """Export channels to a CSV document"""

    selected = resolve_range(channels)

    with output:
        ctx.api.export_channels(output, selected=selected, skip_deleted=skip_deleted)


@channel.command("import")
@click.option("-i", "--input", type=click.File("r"), default=sys.stdin)
@click.option(
    "-s",
    "--sync",
    is_flag=True,
    help="Delete channels from the radio " "that do not exist in input",
)
@click.option(
    "-c",
    "--channels",
    multiple=True,
    help="Specify a single chanel (-c 1) or " "a range of channels (-c 1:10)",
)
@click.option(
    "-I",
    "--ignore-errors",
    "ignore_errors",
    is_flag=True,
    help="Continue to import channels if there is an error",
)
@click.pass_obj
@clear_first
def import_channels(ctx, input, sync, channels, ignore_errors):
    """Import channels from a CSV document"""

    selected = resolve_range(channels)

    with input:
        ctx.api.import_channels(input, selected=selected, ignore_errors=ignore_errors)


@channel.command("delete")
@click.option(
    "-c",
    "--channels",
    multiple=True,
    help="Specify a single chanel (-c 1) or " "a range of channels (-c 1:10)",
)
@click.pass_obj
@clear_first
def delete_channels(ctx, channels):
    """Delete a channel or range of channels"""

    selected = resolve_range(channels)

    for channel in selected:
        LOG.info("deleting channel %d", channel)
        ctx.api.delete_channel_entry(channel)


# ----------------------------------------------------------------------


@main.group()
def info():
    """Commands for getting information about the radio"""
    pass


@info.command("id")
@click.pass_obj
@clear_first
def radio_id(ctx):
    """Return the radio model."""

    res = ctx.api.radio_id()
    print(res)


@info.command("type")
@formatted
@click.pass_obj
@clear_first
def radio_type(ctx):
    """Return the radio type."""

    res = ctx.api.radio_type()
    return res


@info.command()
@formatted
@click.pass_obj
@clear_first
def firmware(ctx):
    """Return information about the radio firmware."""

    res = ctx.api.radio_firmware()
    return res


@info.command()
@formatted
@click.pass_obj
@clear_first
def serial(ctx):
    """Return information about the radio firmware."""

    return ctx.api.radio_serial()


# ----------------------------------------------------------------------


@main.group()
def vfo():
    """Commands for interacting with the VFO"""
    pass


@vfo.command()
@apply_options_from_schema(schema.FO)
@click.argument("band", type=click.Choice(BAND_NAMES))
@formatted
@click.pass_obj
@clear_first
def tune(ctx, band, **kwargs):
    """Get or set VFO frequency and other settings.

    You can only tune to frequencies on the current band. Use the
    frequency-band command to change bands."""

    band = normalize_band(band)
    res = ctx.api.get_band_mode(band)
    if res != 0:
        raise click.ClickException("selected band must be in VFO mode")

    res = ctx.api.get_band_vfo(band)

    set_radio = False
    for k, v in kwargs.items():
        if v is None:
            continue

        set_radio = True
        res[k] = v

    if set_radio:
        LOG.info("setting vfo for band %s", band)
        res = ctx.api.set_band_vfo(band, res)

    return res


@vfo.command("band")
@click.argument("band", type=click.Choice(BAND_NAMES))
@click.argument("freq_band", type=click.Choice(api.FREQUENCY_BAND), required=False)
@click.pass_obj
@clear_first
def frequency_band(ctx, band, freq_band):
    """Get or set the frequency band for the selected radio band.

    Frequency bands are named using the names from "SELECTING A FREQUENCY BAND"
    in the TM-V71 manual. Note that band A and band B support a different
    subset of the available frequencies."""

    band = normalize_band(band)
    with ctx.api.programming_mode():
        if freq_band is None:
            freq_band = ctx.api.get_frequency_band(band)
            pass
        else:
            ctx.api.set_frequency_band(band, freq_band)

    print(freq_band)


# ----------------------------------------------------------------------


@main.group()
def memory():
    """Commands for reading and modifying memory"""
    pass


@memory.command()
@click.option("-o", "--output", type=click.File("wb"), default=sys.stdout.buffer)
@click.pass_obj
@clear_first
def dump(ctx, output):
    """Read entire radio memory and write it to a file."""

    LOG.info('read from radio to file "%s"', output.name)
    try:
        with output, ctx.api.programming_mode():
            try:
                ctx.api.memory_dump(output)
            except api.CommunicationError as err:
                raise click.ClickException(str(err))
    except Exception:
        if output is not sys.stdout.buffer:
            LOG.warning("removing output file %s", output.name)
            os.unlink(output.name)
        raise


@memory.command()
@click.option("-i", "--input", type=click.File("rb"), default=sys.stdin)
@click.pass_obj
@clear_first
def restore(ctx, input):
    """Read memory dump from a file and write it to the radio."""

    LOG.info('write to radio from file "%s"', input.name)
    with input, ctx.api.programming_mode():
        try:
            ctx.api.memory_restore(input)
        except api.CommunicationError as err:
            raise click.ClickException(str(err))


def flexint(v):
    """Convert strings to integer values.

    The block read/write commands accept integer values as either
    decimal or hexadecimal. This function converts either format into
    integer values."""

    if isinstance(v, int):
        return v
    else:
        return int(v, 0)


@memory.command()
@click.option("-o", "--output", type=click.File("wb"), default=sys.stdout.buffer)
@click.option("-h", "--hexdump", "_hexdump", is_flag=True)
@click.option("-l", "--length", type=flexint, default=0)
@click.argument("address")
@click.pass_obj
@clear_first
def read_block(ctx, output, _hexdump, address, length):
    """Read one or more memory blocks from the radio.

    This command will by default output the binary data to stdout. Use the
    '-o' option to write to a file instead. Use the '--hexdump' option
    to output the data as a formatted hexdump instead.

    You can read a range of addresses by specifying the start and end
    (inclusive) of the range separated by a colon.  E.g., to read addresses
    0x1700 through 0x557f, you could use `tmv71 memory read-block
    0x1700:0x557f`.

    Examples:

    1. Read 16 bytes from address 0x1700, version 1:

       tmv71 memory read-block 0x1700 -l 16 -h

    2. Read 16 bytes from address 0x1700, version 2:

       tmv71 memory read-block 0x1700:0x1710 -h
    """

    if ":" in address and length != 0:
        raise ValueError("you cannot specify a length " "when using an address range")
    elif ":" in address:
        addr_start, addr_end = (flexint(x) for x in address.split(":"))
    else:
        addr_start = flexint(address)
        addr_end = addr_start + (length if length else 256)

    with output, ctx.api.programming_mode():
        addr = addr_start
        while addr < addr_end:
            chunklen = min(256, addr_end - addr)

            # The special value '0' means '256 bytes'
            chunklen = 0 if chunklen == 256 else chunklen

            LOG.info(
                "reading %d bytes of memory from 0x%04X",
                256 if chunklen == 0 else chunklen,
                addr,
            )
            data = ctx.api.read_block(addr, chunklen)

            if _hexdump:
                output.write(hexdump.hexdump(data, result="return").encode("ascii"))
                output.write(b"\n")
            else:
                output.write(data)

            addr = min(addr_end, addr + 256)


@memory.command()
@click.option("-i", "--input", type=click.File("rb"))
@click.option("-d", "--hexdata")
@click.option("-s", "--seek", type=flexint, default=0)
@click.option("-l", "--length", type=flexint)
@click.argument("address", type=flexint)
@click.pass_obj
@clear_first
def write_block(ctx, input, hexdata, seek, length, address):
    """Write data to radio memory.

    This command will by default read data from stdin. Use the --input (-i)
    option to read data from a file instead, or --hexdata (-d) to provide
    hexadecimal data on the command line.

    Examples:

    1. Write four bytes to address 0x1700

       tmv71 memory write-block 0x1700 -d 'F0 15 AB 08'

    2. Open 'backup.dat', seek to position 0x1700, read 16 bytes, and
       write them to address 0x1700:

       tmv71 memory write-block 0x1700 -i backup.dat -s 0x1700 -l 16
    """

    LOG.debug("seek %d length %d address %d", seek, length, address)

    if input:
        LOG.debug("reading data from %s", input.name)
        with input:
            input.seek(seek)
            data = input.read(length)
    elif hexdata:
        LOG.debug("reading data from hexdata option")
        data = hexdump.dehex(hexdata)
        if seek:
            data = data[seek:]
        if length:
            data = data[:length]
    elif sys.stdin.buffer.seekable():
        LOG.debug("reading data from stdin (seekable)")
        sys.stdin.buffer.seek(seek)
        data = sys.stdin.buffer.read(length)
    else:
        LOG.debug("reading data from stdin (not seekable)")
        data = sys.stdin.buffer.read()[seek:]
        if length:
            data = data[:length]

    with ctx.api.programming_mode():
        addr = address
        for chunk in hexdump.chunks(data, 256):
            chunklen = len(chunk)
            LOG.info(
                "writing %d bytes of data to address 0x%04X",
                256 if chunklen == 0 else chunklen,
                addr,
            )

            addr += 256
            ctx.api.write_block(address, chunk)


# ----------------------------------------------------------------------


@main.command()
@click.argument("command")
@click.argument("args", nargs=-1)
@click.pass_obj
@clear_first
def raw(ctx, command, args):
    """Send a raw command to the radio and display the response."""

    res = ctx.api.send_command(command, *args)
    print(*res)
